#!/bin/gawk -f
#  
# audit_sudoers.awk v0.4.1 - display /etc/sudoers without any cruft
#  
# Copyright (C) 2014	Henning Rohde (HeRo@amalix.de)
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#  
# TODO:
# - check case-(in)sensitivity
# - "--help"
# - fancier RegEx, esp. regarding "%#gid" in User_Specs or for literal ":" or "#" in commands 
# - aggregate Defaults
# - check "--traditional"-mode
# - eleminate repetitive definitions of an alias
# - ...
# 
########################################
# Change History
# 2017-01-06 v0.4.1 by sandix@universal-logging-system.org 
#                   removed "concatlonglines" with "next" in function
######################################################################

function processUserSuppliedVars( display, showHost, showRunas, showUser,    idx, tmpArr, tmpDispArr ) {
	if ( length( display ) > 0 ) {
		gsub( /[[:space:]]/, "", display );
		split( display, tmpArr, "," )
		for ( idx in tmpArr)
			tmpDispArr[tmpArr[idx]]++;
		if ( "ALL" in tmpDispArr || "Cmnd_Alias" in tmpDispArr  || "Cmnd" in tmpDispArr  || "Command" in tmpDispArr || "Commands" in tmpDispArr )
			Aliases2show["Cmnd_Alias"]++;
		if ( "ALL" in tmpDispArr || "Host_Alias" in tmpDispArr  || "Host" in tmpDispArr  || "Hosts" in tmpDispArr )
			Aliases2show["Host_Alias"]++;
		if ( "ALL" in tmpDispArr || "Runas_Alias" in tmpDispArr || "Runas" in tmpDispArr || "RunAs" in tmpDispArr   || "Run-as" in tmpDispArr || "Run-As" in tmpDispArr || "Run_as" in tmpDispArr || "Run_As" in tmpDispArr )
			Aliases2show["Runas_Alias"]++;
		if ( "ALL" in tmpDispArr || "Tag_Spec" in tmpDispArr    || "Tag" in tmpDispArr   || "Tags" in tmpDispArr )
			Aliases2show["Tag_Spec"]++;
		if ( "ALL" in tmpDispArr || "User_Alias" in tmpDispArr  || "User" in tmpDispArr  || "Users" in tmpDispArr )
			Aliases2show["User_Alias"]++;
	    } else {
		if ( length( showHost ) > 0 && showHost != "ALL" ) {
			gsub( /[[:space:]]/, "", showHost );
			split( showHost, tmpArr, /[[:space:]]*,[[:space:]]*/ )
			for ( idx in tmpArr) Hosts2show[tmpArr[idx]]++;
		    } else {
			split( tolower( ENVIRON["HOSTNAME"] ), tmpArr, "." );
			Hosts2show[tmpArr[1]]++;
		    }
		if ( length( showRunas ) > 0 && showRunas != "ALL" ) {
			gsub( /[[:space:]]/, "", showRunas );
			split( showRunas, tmpArr, /[[:space:]]*,[[:space:]]*/ )
			for ( idx in tmpArr) Runas2show[tmpArr[idx]]++;
		    }
		if ( length( showUser ) > 0 && showUser != "ALL" ) {
			gsub( /[[:space:]]/, "", showUser );
			split( showUser, tmpArr, /[[:space:]]*,[[:space:]]*/ )
			for ( idx in tmpArr) Users2show[tmpArr[idx]]++;
		    }
	    }
    }
function printLabeledArray( prefix, Alias, eq, label2, array2print, width,    idx) {
	printf( "\n# %s %s %s #\n", Alias, eq, label2 );
	for ( idx in array2print ) {
		if ( length( array2print[idx] ) > 0 )
			printf( "%s %-*s%s%s\n", prefix, width, idx, eq, array2print[idx] );
	    }
    }
function add2Cmnd_Alias( alias, values,    idx, valarr) {
# TODO: eleminate repetitive definitions of an alias
#	if ( alias in Cmnd_Alias ) {
#	    } else {
		split( values, valarr, "," );
		for ( idx in valarr ) {
			gsub( /^[[:space:]]+|[[:space:]]+$/, "", valarr[idx] );
# TODO: if allowd: Aliased Command with Tags
			if ( valarr[idx] in Cmnd_Alias && ! ( "Cmnd_Alias" in Aliases2show ) ) {
				add2Cmnd_Alias( alias, Cmnd_Alias[valarr[idx]] );
			    } else {
				if ( length( Aliased_Cmnd[valarr[idx]] ) > 0 )
					Aliased_Cmnd[valarr[idx]] = ( Aliased_Cmnd[valarr[idx]] "," alias );
				    else
					Aliased_Cmnd[valarr[idx]] = alias;
				if ( longest["Cmnd_Alias"] < length( alias ) )
					longest["Cmnd_Alias"] = length( alias );
				if ( length( Cmnd_Alias[alias] ) > 0 )
					Cmnd_Alias[alias] = ( Cmnd_Alias[alias] "," valarr[idx] );
				    else
					Cmnd_Alias[alias] = valarr[idx];
				if ( longest["Aliased_Cmnd"] < length( valarr[idx] ) )
					longest["Aliased_Cmnd"] = length( valarr[idx] );
			    }
		    }
#	    }
    }
function add2Host_Alias( alias, values,    idx, valarr ) {
# TODO: eleminate repetitive definitions of an alias
#	if ( alias in Host_Alias ) {
#	    } else {
		split( values, valarr, "," );
		for ( idx in valarr ) {
			if ( valarr[idx] in Host_Alias && ! ( "Host_Alias" in Aliases2show ) ) {
				add2Host_Alias( alias, Host_Alias[valarr[idx]] );
			    } else {
				if ( length(Aliased_Host[valarr[idx]]) > 0 )
					Aliased_Host[valarr[idx]] = ( Aliased_Host[valarr[idx]] "," alias );
				    else
					Aliased_Host[valarr[idx]] = alias;
				if ( longest["Host_Alias"] < length( alias ) )
					longest["Host_Alias"] = length( alias );
				if ( length(Host_Alias[alias]) > 0 )
					Host_Alias[alias] = ( Host_Alias[alias] "," valarr[idx] );
				    else
					Host_Alias[alias] = valarr[idx];
				if ( longest["Aliased_Host"] < length( valarr[idx] ) )
					longest["Aliased_Host"] = length( valarr[idx] );
			    }
		    }
#	    }
    }
function add2Runas_Alias( alias, values,    idx, valarr) {
# TODO: eleminate repetitive definitions of an alias
#	if ( alias in Runas_Alias ) {
#	    } else {
		split( values, valarr, "," );
		for ( idx in valarr ) {
			if ( valarr[idx] in Runas_Alias && ! ( "Runas_Alias" in Aliases2show ) ) {
				add2Runas_Alias( alias, Runas_Alias[valarr[idx]] );
			    } else {
				if ( length( Aliased_Runas[valarr[idx]] ) > 0 )
					Aliased_Runas[valarr[idx]] = ( Aliased_Runas[valarr[idx]] "," alias );
				    else
					Aliased_Runas[valarr[idx]] = alias;
				if ( longest["Runas_Alias"] < length( alias ) )
					longest["Runas_Alias"] = length( alias );
				if ( length( Runas_Alias[alias] ) > 0 )
					Runas_Alias[alias] = ( Runas_Alias[alias] "," valarr[idx] );
				    else
					Runas_Alias[alias] = valarr[idx];
				if ( longest["Aliased_Runas"] < length( valarr[idx] ) )
					longest["Aliased_Runas"] = length( valarr[idx] );
			    }
		    }
#	    }
    }
function add2User_Alias( alias, values,    idx, valarr) {
# TODO: eleminate repetitive definitions of an alias
#	if ( alias in User_Alias ) {
#	    } else {
		split( values, valarr, "," );
		for ( idx in valarr ) {
			if ( valarr[idx] in User_Alias && ! ( "User_Alias" in Aliases2show ) ) {
				add2User_Alias( alias, User_Alias[valarr[idx]] );
			    } else {
				if ( length(Aliased_User[valarr[idx]]) > 0 )
					Aliased_User[valarr[idx]] = ( Aliased_User[valarr[idx]] "," alias );
				    else
					Aliased_User[valarr[idx]] = alias;
				if ( longest["User_Alias"] < length( alias ) )
					longest["User_Alias"] = length( alias );
				if ( length(User_Alias[alias]) > 0 )
					User_Alias[alias] = ( User_Alias[alias] "," valarr[idx] );
				    else
					User_Alias[alias] = valarr[idx];
				if ( longest["Aliased_User"] < length( valarr[idx] ) )
					longest["Aliased_User"] = length( valarr[idx] );
			    }
		    }
#	    }
    }
function add2Tag_Spec( Tag_List, Cmnd,    Tag_Arr, idx, Cmnd_List ) {
	if ( Cmnd in Cmnd_Alias ) {
		split( Cmnd_Alias[Cmnd], Cmnd_List, /,/ );
		for (idx in Cmnd_List)
			add2Tag_Spec( Tag_List, Cmnd_List[idx] );
	    } else {
		split( Tag_List, Tag_Arr, /:[[:space:]]*/ );
		for ( idx in Tag_Arr ) {
			if ( ( Tag_Arr[idx] ":" ) in Tag_Spec && Tag_Spec[ ( Tag_Arr[idx] ":" ) ] !~ Cmnd ) {
				if ( length( ( Tag_Spec[Tag_Arr[idx] ":"] ) ) > 0 )
					Tag_Spec[ ( Tag_Arr[idx] ":" ) ] = ( Tag_Spec[Tag_Arr[idx] ":"] "," Cmnd );
				    else
					Tag_Spec[ ( Tag_Arr[idx] ":" ) ] = Cmnd;
				if ( longest["Tagged_Cmnd"] < length( Cmnd ) )
					longest["Tagged_Cmnd"] = length( Cmnd );
				if ( length( Tagged_Cmnd[ Cmnd ] ) > 0 )
					Tagged_Cmnd[ Cmnd ] = ( Tagged_Cmnd[ Cmnd ] "," Tag_Arr[idx] ":" );
				    else
					Tagged_Cmnd[ Cmnd ] = ( Tag_Arr[idx] ":" );
				if ( longest["Tag_Spec"] < length( ( Tag_Arr[idx] ":" ) ) )
					longest["Tag_Spec"] = length( ( Tag_Arr[idx] ":" ) );
			    }
		    }
	    }
    }
function ArrNotEmpty( array,    idx ) {
#	isarray( array ) not defined before gawk v4
#	length( array ) not defined in posix awk
	for ( idx in array )
		return 1
	return 0
    }
function add2User_Spec( User_List, Host, Runas, Cmnd,    User, idx ) {
	gsub( /[[:space:]]/, "", User_List );
	split( User_List, User, /,/ );
	for ( idx in User ) {
		if ( User[idx] in User_Alias && ! ( "User_Alias" in Aliases2show ) ) {
			add2User_Spec( User_Alias[User[idx]], Host, Runas, Cmnd);
		    } else
		if (	( ! ArrNotEmpty( Hosts2show )	||	Host in Hosts2show	|| Host == "ALL"	) &&
			( ! ArrNotEmpty( Runas2show )	||	Runas in Runas2show	|| Runas == "ALL"	) &&
			( ! ArrNotEmpty( Users2show )	||	User[idx] in Users2show	|| User[idx] == "ALL"	) ) {
			if ( "User_Alias" in Aliases2show ) {
				if ( longest["User_Alias"] < length( User[idx] ) )
					longest["User_Alias"] = length( User[idx] );
			    } else {
				if ( longest["Aliased_User"] < length( User[idx] ) )
					longest["Aliased_User"] = length( User[idx] );
			    }
			User_Spec_User[ User[idx], Host, Runas, Cmnd ] = User[idx];
			User_Spec_Host[ User[idx], Host, Runas, Cmnd ] = Host;
			User_Spec_Runas[ User[idx], Host, Runas, Cmnd ] = Runas;
			User_Spec_Cmnd[ User[idx], Host, Runas, Cmnd ] = Cmnd;
		    }
	    }
    }
function doRunas_Spec( User_List, Host, Runas_List, Cmnd,    Runas, Runas_Arr, idx ) {
	split( Runas_List, Runas_Arr, /:/);
	if ( length(Runas_Arr[1]) > 0 ) {
		split( Runas_Arr[1], Runas, /,/ );
		for ( idx in Runas ) {
			if ( Runas[idx] in Runas_Alias && ! ( "Runas_Alias" in Aliases2show ) ) {
				doRunas_Spec( User_List, Host, Runas_Alias[Runas[idx]], Cmnd );
			    } else {
				if ( longest["Aliased_Runas"] < length( Runas[idx] ) )
					longest["Aliased_Runas"] = length( Runas[idx] );
				add2User_Spec( User_List, Host, Runas[idx], Cmnd );
			    }
		    }
	    }
	if ( length( Runas_Arr[2] ) > 0 ) {
		split( Runas_Arr[2], Runas, /,/ );
		for ( idx in Runas ) {
			if ( Runas[idx] in Runas_Alias && ! ( "Runas_Alias" in Aliases2show ) ) {
				doRunas_Spec( User_List, Host, ( ":" Runas_Alias[Runas[idx]] ), Cmnd );
			    } else {
				if ( longest["Aliased_Runas"] < length( Runas[idx] ) + 1 )
					longest["Aliased_Runas"] = length( Runas[idx] + 1 );
				add2User_Spec( User_List, Host, ( ":" Runas[idx] ), Cmnd );
			    }
		    }
	    }
    }
function doCmnd_Alias( User_List, Host, Runas_List, Cmnd,    SELinux_Spec, idx, Tag_List, Cmnd_List ) {
	if ( Cmnd ~ /^ROLE=/ || Cmnd ~ /^TYPE=/ ) {
		SELinux_Spec = Cmnd;
		sub( /ROLE=[^[:space:]]+/, "", Cmnd );
		sub( /TYPE=[^[:space:]]+/, "", Cmnd );
		sub( /^[[:space:]]+/, "", Cmnd );
		SELinux_Spec = ( "	" substr( SELinux_Spec, 1, length( SELinux_Spec ) - length( Cmnd ) + 1 ) "	" );
		gsub( /^[[:space:]]+|[[:space:]]+$/, "", SELinux_Spec );
	    }
	if ( match( Cmnd, /[^:]*$/ ) > 1 ) {
		Tag_List = substr( Cmnd, 1, RSTART - 1 );
		sub( /^.*:[[:space:]]*/, "", Cmnd );
		add2Tag_Spec( Tag_List, Cmnd );
	    }
	if ( Cmnd in Cmnd_Alias && ! ( "Cmnd_Alias" in Aliases2show ) ) {
		split( Cmnd_Alias[Cmnd], Cmnd_List, /,/ );
		for ( idx in Cmnd_List ) {
			doRunas_Spec( User_List, Host, Runas_List, ( SELinux_Spec Tag_List Cmnd_List[idx] ) );
		    }
	    } else {
		doRunas_Spec( User_List, Host, Runas_List, ( SELinux_Spec Tag_List Cmnd ) );
	    }
    }
function doCmnd_Spec_List( User_List, Host, Cmnd_Spec_List,    Runas_List, ROL, Cmnd, newCmnd_Spec_List ) {
	if ( match( Cmnd_Spec_List, /\)/ ) > 0 ) {
		Runas_List = substr( Cmnd_Spec_List, 2, RSTART - 2 );
		ROL = substr( Cmnd_Spec_List, RSTART + RLENGTH );
		gsub( /[[:space:]]/, "", Runas_List );
		sub( /^[[:space:]]+/, "", ROL );
		if ( match(ROL, /,/) > 0 ) {
# TODO: Smarter RegEx needed: What if command contains a comma?
			Cmnd = substr( ROL, 1, RSTART - 1 );
			newCmnd_Spec_List = substr( ROL, RSTART + RLENGTH );
			sub( /^[[:space:]]+/, "", newCmnd_Spec_List );
			doCmnd_Alias( User_List, Host, Runas_List, Cmnd );
			doCmnd_Spec_List( User_List, Host, "(" Runas_List ")" newCmnd_Spec_List );
		    } else {
			doCmnd_Alias( User_List, Host, Runas_List, ROL );
		    }
	    } else {
		doCmnd_Spec_List( User_List, Host, ( "(root)" Cmnd_Spec_List ) );
	    }
    }
function doHost_List( User_List, Host_List, Cmnd_Spec_List,    Host, idx ) {
	gsub( /[[:space:]]/, "", Host_List );
	split( Host_List, Host, /,/ );
	for ( idx in Host ) {
		if ( Host[idx] in Host_Alias && ! ( "Host_Alias" in Aliases2show ) ) {
			doHost_List( User_List, Host_Alias[Host[idx]], Cmnd_Spec_List );
		    } else {
			if ( longest["Aliased_Host"] < length( Host[idx] ) )
				longest["Aliased_Host"] = length( Host[idx] );
			doCmnd_Spec_List( User_List, Host[idx], Cmnd_Spec_List );
		    }
	    }
    }
function doHLeqCSL( User_List, HLeqCSL,    Cmnd_Spec_List, Host_List ) {
	Host_List = substr( HLeqCSL, 1, match( HLeqCSL, /[[:space:]]*=[[:space:]]*/ ) - 1 );
	Cmnd_Spec_List = substr( HLeqCSL, RSTART + RLENGTH );
	sub( /^[[:space:]]*=[[:space:]]*/, "", Cmnd_Spec_List );
	doHost_List( User_List, Host_List, Cmnd_Spec_List );
    }
function doHLeqCSLlist( User_List, HLeqCSLlist,    HLeqCSL ) {
# TODO: Smarter RegEx needed: What if command contains a colon?
	if ( match( HLeqCSLlist, /:[,[:alpha:][:digit:][:space:]]+=/ )==0 ) {
		doHLeqCSL( User_List, HLeqCSLlist );
	    } else {
		HLeqCSL = substr( HLeqCSLlist, 1, RSTART - 1 );
		HLeqCSLlist = substr( HLeqCSLlist, RSTART + 1 );
		sub( /[[:space:]]+$/, "", HLeqCSL );
		sub( /^[[:space:]]+/, "", HLeqCSLlist );
		doHLeqCSL( User_List, HLeqCSL );
		doHLeqCSLlist( User_List, HLeqCSLlist );
	    }
    }
function doLines( line,    linearr, idx, alias, values ) {
	if ( line ~ /^Defaults/ ) {
# TODO: Aggregate Defaults
	    } else
	if ( line ~ /^Cmnd_Alias/ ) {
		gsub( /^Cmnd_Alias[[:space:]]+|[[:space:]]+$/, "", line );
# TODO: Smarter RegEx needed: What if command contains a colon?
		split( line, linearr, /[[:space:]]*:[[:space:]]*/ );
		for ( idx in linearr ) {
			alias = substr( linearr[idx], 1, match( linearr[idx], /[[:space:]]*=/ )-1 );
			values = substr( linearr[idx], RSTART + RLENGTH );
			add2Cmnd_Alias( alias, values );
		    }
	    } else
	if ( line ~ /^Host_Alias/ ) {
		gsub( /^Host_Alias[[:space:]]+|[[:space:]]/, "", line );
		split( line, linearr, /:/ );
		for ( idx in linearr ) {
			alias = substr( linearr[idx], 1, match( linearr[idx], /[[:space:]]*=/ )-1 );
			values = substr( linearr[idx], RSTART + RLENGTH );
			add2Host_Alias( alias, values );
		    }
	    } else
	if ( line ~ /^Runas_Alias/ ) {
		gsub( /^Runas_Alias[[:space:]]+|[[:space:]]/, "", line );
		split( line, linearr, /:/ );
		for ( idx in linearr ) {
			alias = substr( linearr[idx], 1, match( linearr[idx], /[[:space:]]*=/ )-1 );
			values = substr( linearr[idx], RSTART + RLENGTH );
			add2Runas_Alias( alias, values );
		    }
	    } else
	if ( line ~ /^User_Alias/ ) {
		gsub( /^User_Alias[[:space:]]+|[[:space:]]/, "", line );
# TODO: Smarter RegEx needed: What if #uid or #gid is used?
		split( line, linearr, /:/ );
		for ( idx in linearr ) {
			alias = substr( linearr[idx], 1, match( linearr[idx], /=/ )-1 );
			values = substr( linearr[idx], RSTART + RLENGTH );
			add2User_Alias( alias, values );
		    }
	    } else
	{
# _User Specification_
# User_Spec ::= User_List Host_List '=' Cmnd_Spec_List (':' Host_List '=' Cmnd_Spec_List)*
# Cmnd_Spec_List ::= Cmnd_Spec | Cmnd_Spec ',' Cmnd_Spec_List
# Cmnd_Spec ::= Runas_Spec? SELinux_Spec? Tag_Spec* Cmnd
# Runas_Spec ::= '(' Runas_List? (':' Runas_List)? ')'
# SELinux_Spec ::= ('ROLE=role' | 'TYPE=type')
# Tag_Spec ::= ('NOPASSWD:' | 'PASSWD:' | 'NOEXEC:' | 'EXEC:' | 'SETENV:' | 'NOSETENV:' | 'LOG_INPUT:' | 'NOLOG_INPUT:' | 'LOG_OUTPUT:' | 'NOLOG_OUTPUT:')
# A user specification determines which commands a user may run (and as what user) on specified hosts.  By default, commands are run as root, but this can be changed on a per-command basis.

		User_List = substr( line, 1, match( line, /[^[:space:],][[:space:]]+[^[:space:],]/ ) );
		values = substr( line, RSTART + RLENGTH - 1 );
		sub( /^[[:space:]]+/, "", values );

		doHLeqCSLlist( User_List, values );
	    }
    }

function processFile( file ) {
    prevLine = ""
	while( (getline newLine < file) > 0 ) {
		while ( match( newLine, /\\$/) ) {
			gsub( /^[[:space:]]+|[[:space:]]+$/, "	", newLine );
			prevLine = prevLine substr(newLine, 1, RSTART - 1)
			if( (getline newLine < file) <= 0 )
				break
		}
		gsub( /^[[:space:]]+|[[:space:]]+$/, "	", newLine );
		line = prevLine newLine
		prevLine = ""
		if ( match( line, /^#includedir[[:space:]]+/ ) ) {
			rdir = substr( line, RSTART + RLENGTH );
			sub( /#.*$/, "", rdir );
			rfiles = "find " rdir " -follow -type f"
			while( ( rfiles | getline rfile ) > 0 )
				if ( files_read[rfile]++ == 0 )
					if( ifile != "" )
						processFile(ifile)
			close( rfiles );
			line = ""
		    }
		else if ( match( line, /^#include[[:space:]]+/ ) ) {
			ifile = substr( line, RSTART + RLENGTH );
			sub( /#.*$/, "", ifile );
			if ( files_read[ifile]++ == 0 )
				if( ifile != "" )
					processFile(ifile)
			line = ""
		    }
		else if ( line !~ /^#|^$/ ) {
# TODO: Smarter RegEx needed: What if #uid or %#gid is used?
			if ( match( line, /[^\\]#/ ) > 0 ) {
				line = substr( line, 1, RSTART );
				gsub( /[[:space:]]+$/, "", nextLine );
				}
			doLines( line );
			}
	    	}
	}

function processMain() {
    prevLine = ""
	while( getline newLine > 0 ) {
		while ( match( newLine, /\\$/) ) {
			gsub( /^[[:space:]]+|[[:space:]]+$/, "	", newLine );
			prevLine = prevLine substr(newLine, 1, RSTART - 1)
			if( getline newLine <= 0 )
				break
		}
		gsub( /^[[:space:]]+|[[:space:]]+$/, "	", newLine );
		line = prevLine newLine
		prevLine = ""
		if ( match( line, /^#includedir[[:space:]]+/ ) ) {
			rdir = substr( line, RSTART + RLENGTH );
			sub( /#.*$/, "", rdir );
			rfiles = "find " rdir " -follow -type f"
			while( ( rfiles | getline rfile ) > 0 )
				if ( files_read[rfile]++ == 0 )
					if( rfile != "" )
						processFile(rfile)
			close( rfiles );
			line = ""
		    }
		else if ( match( line, /^#include[[:space:]]+/ ) ) {
			ifile = substr( line, RSTART + RLENGTH );
			sub( /#.*$/, "", ifile );
			if ( files_read[ifile]++ == 0 )
				if( ifile != "" )
					processFile(ifile)
			line = ""
		    }
		else if ( line !~ /^#|^$/ ) {
# TODO: Smarter RegEx needed: What if #uid or %#gid is used?
			if ( match( line, /[^\\]#/ ) > 0 ) {
				line = substr( line, 1, RSTART );
				gsub( /[[:space:]]+$/, "", nextLine );
				}
			doLines( line );
			}
    		}
	}

BEGIN{
	split( "", Aliases2show );
	split( "", Hosts2show );
	split( "", Runas2show );
	split( "", Users2show );
	split( "", Aliased_Cmnd );
	split( "", Aliased_User );
	split( "", Cmnd_Alias );
	split( "", Tagged_Cmnd );
	split( "", User_Alias );
	split( "", User_Spec_User );
	split( "", User_Spec_Host );
	split( "", User_Spec_Runas );
	split( "", User_Spec_Cmnd );
	split( "", files_read );
	split( "", longest );
	processUserSuppliedVars( display, showHost, showRunas, showUser );

	Tag_Spec["NOPASSWD:"]="";
	Tag_Spec["PASSWD:"]="";
	Tag_Spec["NOEXEC:"]="";
	Tag_Spec["EXEC:"]="";
	Tag_Spec["SETENV:"]="";
	Tag_Spec["NOSETENV:"]="";
	Tag_Spec["LOG_INPUT:"]="";
	Tag_Spec["NOLOG_INPUT:"]="";
	Tag_Spec["LOG_OUTPUT:"]="";
	Tag_Spec["NOLOG_OUTPUT:"]="";

	processMain() 
    }

END {
	PROCINFO["sorted_in"] = "@ind_str_desc"
	if ( "Cmnd_Alias" in Aliases2show && ( ArrNotEmpty( Cmnd_Alias ) ) ) {
		printLabeledArray( "Cmnd_Alias", "Cmnd_Alias", " = ", "Commands", Cmnd_Alias, longest["Cmnd_Alias"] );
		printLabeledArray( "# ", "Command", "Cmnd_Alias", " in ", Aliased_Cmnd, longest["Aliased_Cmnd"] );
	    }
	if ( "Host_Alias" in Aliases2show && ( ArrNotEmpty( Host_Alias ) ) ) {
		printLabeledArray( "Host_Alias", "Host_Alias", " = ", "Hosts", Host_Alias, longest["Host_Alias"] );
		printLabeledArray( "# ", "Host", "Host_Alias", " in ", Aliased_Host, longest["Aliased_Host"] );
	    }
	if ( "Runas_Alias" in Aliases2show && ( ArrNotEmpty( Runas_Alias ) ) ) {
		printLabeledArray( "Runas_Alias", "Runas_Alias", " = ", "Runas-User/-Groups", Runas_Alias, longest["Runas_Alias"] );
		printLabeledArray( "# ", "Runas-User/-Group", " in ", "Runas_Alias", Aliased_Runas, longest["Aliased_Runas"] );
	    }
	if ( "User_Alias" in Aliases2show && ( ArrNotEmpty( User_Alias ) ) ) {
		printLabeledArray( "User_Alias", "User_Alias", " = ", "User/Groups", User_Alias, longest["User_Alias"] );
		printLabeledArray( "# ", "User/Group", " in ", "User_Alias", Aliased_User, longest["Aliased_User"] );
	    }
	if ( "Tag_Spec" in Aliases2show && ( ArrNotEmpty( Tagged_Cmnd ) ) ) {
		printLabeledArray( "# ", "Tag", " on ", "Commands", Tag_Spec );
		printLabeledArray( "# ", "Command", " with ", "Tags", Tagged_Cmnd, longest["Tagged_Cmnd"] );
	    }
	printf( "\n# User_Specs #\n" );
	for ( idx in User_Spec_User ) {
		if ( "User_Alias" in Aliases2show )
			printf( "%-*s %*s = ( %-*s )%s\n", longest["User_Alias"], User_Spec_User[idx], longest["Aliased_Host"], User_Spec_Host[idx], longest["Aliased_Runas"], User_Spec_Runas[idx], User_Spec_Cmnd[idx] );
		    else
			printf( "%-*s %*s = ( %-*s )%s\n", longest["Aliased_User"], User_Spec_User[idx], longest["Aliased_Host"], User_Spec_Host[idx], longest["Aliased_Runas"], User_Spec_Runas[idx], User_Spec_Cmnd[idx] );
	    }
}
